@@ -13,19 +13,24 @@ class Flashcart:
 	CART_READ_FNCPTR = None
 	CART_POWERCYCLE_FNCPTR = None
 	PROGRESS_FNCPTR = None
+	SET_WE_PIN_WR = None
+	SET_WE_PIN_AUDIO = None
+	DEFAULT_WE = None
 	SECTOR_COUNT = 0
 	SECTOR_POS = 0
 	SECTOR_MAP = None
 	CFI = None
 	LAST_SR = 0x00
 
-	def __init__(self, config=None, cart_write_fncptr=None, cart_write_fast_fncptr=None, cart_read_fncptr=None, cart_powercycle_fncptr=None, progress_fncptr=None):
+	def __init__(self, config=None, cart_write_fncptr=None, cart_write_fast_fncptr=None, cart_read_fncptr=None, cart_powercycle_fncptr=None, progress_fncptr=None, set_we_pin_wr=None, set_we_pin_audio=None):
 		if config is None: config = {}
 		self.CART_WRITE_FNCPTR = cart_write_fncptr
 		self.CART_WRITE_FAST_FNCPTR = cart_write_fast_fncptr
 		self.CART_READ_FNCPTR = cart_read_fncptr
 		self.CART_POWERCYCLE_FNCPTR = cart_powercycle_fncptr
 		self.PROGRESS_FNCPTR = progress_fncptr
+		self.SET_WE_PIN_WR = set_we_pin_wr
+		self.SET_WE_PIN_AUDIO = set_we_pin_audio
 		self.CONFIG = config
 		if "command_set" in config:
 			self.CONFIG["_command_set"] = config["command_set"]
@@ -33,6 +38,8 @@ class Flashcart:
 			self.CONFIG["_command_set"] = "INTEL"
 		else:
 			self.CONFIG["_command_set"] = ""
+		if "write_pin" in config:
+			self.DEFAULT_WE = config["write_pin"]
 	
 	def CartRead(self, address, length=0):
 		if length == 0:
@@ -42,16 +49,16 @@ class Flashcart:
 				length = 1
 		return self.CART_READ_FNCPTR(address, length)
 	
-	def CartWrite(self, commands, flashcart=True, sram=False):
-		if "command_set" in self.CONFIG and self.CONFIG["command_set"] in ("GBMEMORY", "DMG-MBC5-32M-FLASH"): flashcart = False
-		dprint(commands, flashcart, sram)
-		if flashcart and not sram:
+	def CartWrite(self, commands, fast_write=True, sram=False):
+		if "command_set" in self.CONFIG and self.CONFIG["command_set"] in ("GBMEMORY", "DMG-MBC5-32M-FLASH"): fast_write = False
+		dprint(commands, fast_write, sram)
+		if fast_write and not sram:
 			self.CART_WRITE_FAST_FNCPTR(commands, flashcart=True)
 		else:
 			for command in commands:
 				address = command[0]
 				value = command[1]
-				self.CART_WRITE_FNCPTR(address, value, flashcart=flashcart, sram=sram)
+				self.CART_WRITE_FNCPTR(address, value, flashcart=fast_write, sram=sram)
 	
 	def GetCommandSetType(self):
 		return self.CONFIG["_command_set"].upper()
@@ -64,7 +71,7 @@ class Flashcart:
 
 	def GetVoltage(self):
 		return self.CONFIG["voltage"]
-
+	
 	def GetMBC(self):
 		if (self.CONFIG["type"].upper() == "AGB") or ("mbc" not in self.CONFIG): return False
 		mbc = self.CONFIG["mbc"]
@@ -158,7 +165,7 @@ class Flashcart:
 					dprint("Reading 0x{:X} bytes from cartridge at 0x{:X} = {:s}".format(command[1], command[0], str(temp)))
 			time.sleep(0.001)
 		if "unlock" in self.CONFIG["commands"]:
-			self.CartWrite(self.CONFIG["commands"]["unlock"])
+			self.CartWrite(self.CONFIG["commands"]["unlock"], fast_write=False)
 			time.sleep(0.001)
 
 	def Reset(self, full_reset=False, max_address=0x2000000):
@@ -172,15 +179,15 @@ class Flashcart:
 				if j >= max_address: break
 				dprint("reset_every @ 0x{:X}".format(j))
 				for command in self.CONFIG["commands"]["reset"]:
-					self.CartWrite([[j, command[1]]])
+					self.CartWrite([[j + command[0], command[1]]])
 					time.sleep(0.01)
 		elif "reset" in self.CONFIG["commands"]:
 			self.CartWrite(self.CONFIG["commands"]["reset"])
 			time.sleep(0.001)
 	
 	def VerifyFlashID(self):
-		if "read_identifier" not in self.CONFIG["commands"]: return False
-		if len(self.CONFIG["flash_ids"]) == 0: return False
+		if "read_identifier" not in self.CONFIG["commands"]: return (False, [])
+		if len(self.CONFIG["flash_ids"]) == 0: return (False, [])
 		if "power_cycle" in self.CONFIG and self.CONFIG["power_cycle"] is True:
 			self.CART_POWERCYCLE_FNCPTR()
 		self.Reset()
@@ -233,6 +240,27 @@ class Flashcart:
 			smallest_sector_size = min(smallest_sector_size, sector[0])
 		return smallest_sector_size
 	
+	def GetSectorOffsets(self, rom_size=0, rom_bank_size=0x4000):
+		regions = self.GetSectorMap()
+		pos = 0
+		offsets = []
+		if isinstance(regions, list):
+			for region in regions:
+				size = region[0]
+				count = region[1]
+				for _ in range(0, count):
+					# offsets.append([ pos, math.floor(pos / rom_bank_size) ])
+					offsets.append([ pos, size ])
+					pos += size
+			# for i in range(0, len(offsets)):
+			# 	#print(hex(offsets[i][0]), hex(offsets[i][1]))
+			# 	print(hex(offsets[i]))
+		else:
+			while pos < rom_size:
+				offsets.append([ pos, regions ])
+				pos += regions
+		return offsets
+	
 	def GetSectorMap(self):
 		if self.SECTOR_MAP is not None:
 			return self.SECTOR_MAP
@@ -264,8 +292,23 @@ class Flashcart:
 		for i in range(0, len(self.CONFIG["commands"]["chip_erase"])):
 			addr = self.CONFIG["commands"]["chip_erase"][i][0]
 			data = self.CONFIG["commands"]["chip_erase"][i][1]
+			if len(self.CONFIG["commands"]["chip_erase"][i]) > 2:
+				we = self.CONFIG["commands"]["chip_erase"][i][2]
+			else:
+				we = None
+			
 			if not addr == None:
+				if we == "WR":
+					self.SET_WE_PIN_WR()
+				elif we == "AUDIO":
+					self.SET_WE_PIN_AUDIO()
 				self.CartWrite([[addr, data]])
+				if we is not None:
+					if self.DEFAULT_WE == "WR":
+						self.SET_WE_PIN_WR()
+					elif self.DEFAULT_WE == "AUDIO":
+						self.SET_WE_PIN_AUDIO()
+			
 			time.sleep(0.1)
 			if self.CONFIG["commands"]["chip_erase_wait_for"][i][0] != None:
 				addr = self.CONFIG["commands"]["chip_erase_wait_for"][i][0]
@@ -277,7 +320,18 @@ class Flashcart:
 						for j in range(0, len(self.CONFIG["commands"]["read_status_register"])):
 							#sr_addr = self.CONFIG["commands"]["read_status_register"][j][0]
 							sr_data = self.CONFIG["commands"]["read_status_register"][j][1]
+							
+							if we == "WR":
+								self.SET_WE_PIN_WR()
+							elif we == "AUDIO":
+								self.SET_WE_PIN_AUDIO()
 							self.CartWrite([[addr, sr_data]])
+							if we is not None:
+								if self.DEFAULT_WE == "WR":
+									self.SET_WE_PIN_WR()
+								elif self.DEFAULT_WE == "AUDIO":
+									self.SET_WE_PIN_AUDIO()
+
 					self.CartRead(addr, 2) # dummy read (fixes some bootlegs)
 					wait_for = struct.unpack("<H", self.CartRead(addr, 2))[0]
 					self.LAST_SR = wait_for
@@ -299,13 +353,28 @@ class Flashcart:
 		for i in range(0, len(self.CONFIG["commands"]["sector_erase"])):
 			addr = self.CONFIG["commands"]["sector_erase"][i][0]
 			data = self.CONFIG["commands"]["sector_erase"][i][1]
+			if len(self.CONFIG["commands"]["sector_erase"][i]) > 2:
+				we = self.CONFIG["commands"]["sector_erase"][i][2]
+			else:
+				we = None
+			
 			if addr == "SA": addr = pos
 			if addr == "SA+1": addr = pos + 1
 			if addr == "SA+2": addr = pos + 2
 			if addr == "SA+0x4000": addr = pos + 0x4000
 			if addr == "SA+0x7000": addr = pos + 0x7000
 			if not addr == None:
+				if we == "WR":
+					self.SET_WE_PIN_WR()
+				elif we == "AUDIO":
+					self.SET_WE_PIN_AUDIO()
 				self.CartWrite([[addr, data]])
+				if we is not None:
+					if self.DEFAULT_WE == "WR":
+						self.SET_WE_PIN_WR()
+					elif self.DEFAULT_WE == "AUDIO":
+						self.SET_WE_PIN_AUDIO()
+			
 			if self.CONFIG["commands"]["sector_erase_wait_for"][i][0] != None:
 				addr = self.CONFIG["commands"]["sector_erase_wait_for"][i][0]
 				data = self.CONFIG["commands"]["sector_erase_wait_for"][i][1]
@@ -321,8 +390,24 @@ class Flashcart:
 						for j in range(0, len(self.CONFIG["commands"]["read_status_register"])):
 							sr_addr = self.CONFIG["commands"]["read_status_register"][j][0]
 							sr_data = self.CONFIG["commands"]["read_status_register"][j][1]
+
+							if we == "WR":
+								self.SET_WE_PIN_WR()
+							elif we == "AUDIO":
+								self.SET_WE_PIN_AUDIO()
 							self.CartWrite([[sr_addr, sr_data]])
+							if we is not None:
+								if self.DEFAULT_WE == "WR":
+									self.SET_WE_PIN_WR()
+								elif self.DEFAULT_WE == "AUDIO":
+									self.SET_WE_PIN_AUDIO()
+							
 					self.CartRead(addr, 2) # dummy read (fixes some bootlegs)
+					temp = self.CartRead(addr, 2)
+					if len(temp) != 2:
+						dprint("Communication error in SectorErase():", temp)
+						#self.PROGRESS_FNCPTR({"info_type":"msgbox_critical", "info_msg":"A critical communication error occured during sector erase. Please avoid passive USB hubs, try different USB ports/cables and re-connect the device."})
+						return False
 					wait_for = struct.unpack("<H", self.CartRead(addr, 2))[0]
 					self.LAST_SR = wait_for
 					dprint("Status Register Check: 0x{:X} & 0x{:X} == 0x{:X}? {:s}".format(wait_for, self.CONFIG["commands"]["sector_erase_wait_for"][i][2], data, str(wait_for & self.CONFIG["commands"]["sector_erase_wait_for"][i][2] == data)))
@@ -330,8 +415,8 @@ class Flashcart:
 					time.sleep(0.1)
 					timeout -= 1
 					if timeout < 1:
-						dprint("Timeout error!")
-						self.PROGRESS_FNCPTR({"action":"ABORT", "info_type":"msgbox_critical", "info_msg":"The sector erase attempt timed out. The last status register value was 0x{:X}.\n\nPlease make sure that the cartridge contacts are clean, and that the selected cartridge type and settings are correct.".format(self.LAST_SR), "abortable":False})
+						dprint("Timeout error in SectorErase():", self.LAST_SR)
+						#self.PROGRESS_FNCPTR({"action":"ABORT", "info_type":"msgbox_critical", "info_msg":"The sector erase attempt timed out. The last status register value was 0x{:X}.\n\nPlease make sure that the cartridge contacts are clean, and that the selected cartridge type and settings are correct.".format(self.LAST_SR), "abortable":False})
 						return False
 					if wait_for == data: break
 					self.PROGRESS_FNCPTR({"action":"SECTOR_ERASE", "sector_pos":buffer_pos, "time_start":time.time(), "abortable":True})
